package de.hypoport.emma.nl.rules.platform.calculations;

import org.apache.log4j.Logger;
import org.drools.KnowledgeBase;
import org.drools.KnowledgeBaseConfiguration;
import org.drools.KnowledgeBaseFactory;
import org.drools.builder.KnowledgeBuilder;
import org.drools.builder.KnowledgeBuilderError;
import org.drools.builder.KnowledgeBuilderErrors;
import org.drools.builder.KnowledgeBuilderFactory;
import org.drools.builder.ResourceType;
import org.drools.io.ResourceFactory;
import org.testng.TestException;
import org.testng.annotations.DataProvider;

import java.io.File;
import java.lang.reflect.Method;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;

import static org.fest.assertions.Assertions.assertThat;

public class KnowledgeBaseDataProvider {

  private static Logger log = Logger.getLogger(KnowledgeBaseDataProvider.class);

  @DataProvider
  public static Object[][] compileRuleFiles(Method testMethod) {
    final RuleFiles ruleFileAnnotation = testMethod.getAnnotation(RuleFiles.class);
    if (ruleFileAnnotation == null) {
      throw new TestException("Please specify @RuleFiles for test method " + testMethod.getName());
    }
    return new Object[][] {packageRuleFilesForDebuggingOfBosScoring(ruleFileAnnotation.value())};
  }

  private static Object[] packageRuleFilesForDebuggingOfBosScoring(String[] filesToPackage) {
    KnowledgeBaseConfiguration configuration = KnowledgeBaseFactory.newKnowledgeBaseConfiguration();
    KnowledgeBase knowledgeBase = KnowledgeBaseFactory.newKnowledgeBase(configuration);
    KnowledgeBuilder builder = KnowledgeBuilderFactory.newKnowledgeBuilder();

    for (String currentRuleFile : filesToPackage) {
      File rulesFile = new File(currentRuleFile);
      if (!rulesFile.exists()) {
        log.error("Rule file " + rulesFile.getAbsolutePath() + " must exist.");
      }

      log.debug("Adding rule file " + rulesFile.getPath() + "...");
      builder.add(ResourceFactory.newFileResource(rulesFile), ResourceType.DRL);
      handleBuilderErrors(builder, currentRuleFile);
      log.debug("File added successfully.");
    }
    log.debug("All rule files added successfully.");
    knowledgeBase.addKnowledgePackages(builder.getKnowledgePackages());
    return new Object[] {knowledgeBase, filesToPackage};
  }

  private static void handleBuilderErrors(KnowledgeBuilder build, String fileName) {
    if (build.hasErrors()) {
      StringBuffer allErrors = new StringBuffer();
      allErrors.append("Error compiling DRL file: ");
      allErrors.append(fileName);
      allErrors.append('\n');
      KnowledgeBuilderErrors knowledgeBuilderErrors = build.getErrors();
      for (KnowledgeBuilderError knowledgeBuilderError : knowledgeBuilderErrors) {
        String message = "At [" + formatLineInformation(knowledgeBuilderError.getErrorLines()) + "]: "
                         + knowledgeBuilderError.getMessage();
        allErrors.append(message).append("\n");
      }
      log.error(allErrors.toString());
      throw new RuntimeException(allErrors.toString());
    }
  }

  private static String formatLineInformation(int[] errorLines) {
    StringBuilder stringBuilder = new StringBuilder();
    for (int line : errorLines) {
      stringBuilder.append(line);
      stringBuilder.append(",");
    }
    return stringBuilder.toString();
  }
}
