package de.hypoport.emma.nl.rules.platform.calculations;

import org.apache.log4j.Logger;
import org.drools.KnowledgeBase;
import org.drools.ObjectFilter;
import org.drools.definition.type.FactType;
import org.drools.impl.KnowledgeBaseImpl;
import org.drools.reteoo.Rete;
import org.drools.reteoo.ReteooRuleBase;
import org.drools.rule.Package;
import org.drools.runtime.StatefulKnowledgeSession;
import org.testng.annotations.Test;

import java.util.Collection;

import static org.fest.assertions.Assertions.assertThat;

@Test
public class RuleDoesNotFireIfUsingDeclaredFunctionAndTooManyPackages {

  private static Logger log = Logger.getLogger(RuleDoesNotFireIfUsingDeclaredFunctionAndTooManyPackages.class);
  private static final String REGULAR_PACKAGE_NAME = "de.hypoport.emma.nl.rules.platform.calculations";
  private static final String SLIGHTLY_DIFFERENT_PACKAGE_NAME = "de.hypoport.emma.nl.rules.platform.calculations2";

  @Test(dataProviderClass = KnowledgeBaseDataProvider.class, dataProvider = "compileRuleFiles")
  @RuleFiles(
      {
          "src/test/rules/de/hypoport/emma/nl/rules/platform/calculations/declared-type.drl",
          "src/test/rules/de/hypoport/emma/nl/rules/platform/calculations/check-rule-for-test.drl"
      }
  )
  public void with_only_two_rule_files_everything_is_fine(KnowledgeBase knowledgeBase, String[] packagedRuleFiles) throws Exception {
    runTestForRuleSet(knowledgeBase, packagedRuleFiles, REGULAR_PACKAGE_NAME);
  }

  @Test(dataProviderClass = KnowledgeBaseDataProvider.class, dataProvider = "compileRuleFiles")
  @RuleFiles(
      {
          "src/test/rules/de/hypoport/emma/nl/rules/platform/calculations/declared-type.drl",
          "src/test/rules/de/hypoport/emma/nl/rules/platform/calculations/check-rule-for-test-in-slightly-different-package.drl"
      }
  )
  public void when_renaming_the_package_of_the_check_rule_the_rule_should_still_fire(KnowledgeBase knowledgeBase, String[] packagedRuleFiles) throws Exception {
    runTestForRuleSet(knowledgeBase, packagedRuleFiles, REGULAR_PACKAGE_NAME);
  }


  @Test(dataProviderClass = KnowledgeBaseDataProvider.class, dataProvider = "compileRuleFiles")
  @RuleFiles(
      {
          "src/test/rules/de/hypoport/emma/nl/rules/platform/calculations/declared-type.drl",
          "src/test/rules/de/hypoport/emma/nl/rules/platform/calculations/check-rule-for-test.drl",
          "src/test/rules/de/hypoport/emma/nl/rules/simple-rule-100.drl",
          "src/test/rules/de/hypoport/emma/nl/rules/simple-rule-101.drl",
          "src/test/rules/de/hypoport/emma/nl/rules/simple-rule-102.drl",
          "src/test/rules/de/hypoport/emma/nl/rules/simple-rule-103.drl",
          "src/test/rules/de/hypoport/emma/nl/rules/simple-rule-104.drl",
          "src/test/rules/de/hypoport/emma/nl/rules/simple-rule-105.drl",
          "src/test/rules/de/hypoport/emma/nl/rules/simple-rule-106.drl",
          "src/test/rules/de/hypoport/emma/nl/rules/simple-rule-107.drl",
          "src/test/rules/de/hypoport/emma/nl/rules/simple-rule-108.drl",
          "src/test/rules/de/hypoport/emma/nl/rules/simple-rule-109.drl",
          "src/test/rules/de/hypoport/emma/nl/rules/simple-rule-110.drl",
          "src/test/rules/de/hypoport/emma/nl/rules/simple-rule-111.drl",
          "src/test/rules/de/hypoport/emma/nl/rules/simple-rule-112.drl",
          "src/test/rules/de/hypoport/emma/nl/rules/simple-rule-113.drl",
          "src/test/rules/de/hypoport/emma/nl/rules/simple-rule-114.drl",
          "src/test/rules/de/hypoport/emma/nl/rules/simple-rule-115.drl",
          "src/test/rules/de/hypoport/emma/nl/rules/simple-rule-116.drl",
          "src/test/rules/de/hypoport/emma/nl/rules/simple-rule-117.drl",
          "src/test/rules/de/hypoport/emma/nl/rules/simple-rule-118.drl",
          "src/test/rules/de/hypoport/emma/nl/rules/simple-rule-119.drl",
          "src/test/rules/de/hypoport/emma/nl/rules/simple-rule-120.drl",
          "src/test/rules/de/hypoport/emma/nl/rules/simple-rule-121.drl"
      }
  )
  public void with_two_rule_files_plus_a_small_number_of_dummy_packages_it_is_still_fine(KnowledgeBase knowledgeBase, String[] packagedRuleFiles) throws Exception {
    runTestForRuleSet(knowledgeBase, packagedRuleFiles, REGULAR_PACKAGE_NAME);
  }

  @Test(dataProviderClass = KnowledgeBaseDataProvider.class, dataProvider = "compileRuleFiles")
  @RuleFiles(
      {
          "src/test/rules/de/hypoport/emma/nl/rules/platform/calculations/declared-type.drl",
          "src/test/rules/de/hypoport/emma/nl/rules/platform/calculations/check-rule-for-test.drl",
          "src/test/rules/de/hypoport/emma/nl/rules/simple-rule-100.drl",
          "src/test/rules/de/hypoport/emma/nl/rules/simple-rule-101.drl",
          "src/test/rules/de/hypoport/emma/nl/rules/simple-rule-102.drl",
          "src/test/rules/de/hypoport/emma/nl/rules/simple-rule-103.drl",
          "src/test/rules/de/hypoport/emma/nl/rules/simple-rule-104.drl",
          "src/test/rules/de/hypoport/emma/nl/rules/simple-rule-105.drl",
          "src/test/rules/de/hypoport/emma/nl/rules/simple-rule-106.drl",
          "src/test/rules/de/hypoport/emma/nl/rules/simple-rule-107.drl",
          "src/test/rules/de/hypoport/emma/nl/rules/simple-rule-108.drl",
          "src/test/rules/de/hypoport/emma/nl/rules/simple-rule-109.drl",
          "src/test/rules/de/hypoport/emma/nl/rules/simple-rule-110.drl",
          "src/test/rules/de/hypoport/emma/nl/rules/simple-rule-111.drl",
          "src/test/rules/de/hypoport/emma/nl/rules/simple-rule-112.drl",
          "src/test/rules/de/hypoport/emma/nl/rules/simple-rule-113.drl",
          "src/test/rules/de/hypoport/emma/nl/rules/simple-rule-114.drl",
          "src/test/rules/de/hypoport/emma/nl/rules/simple-rule-115.drl",
          "src/test/rules/de/hypoport/emma/nl/rules/simple-rule-116.drl",
          "src/test/rules/de/hypoport/emma/nl/rules/simple-rule-117.drl",
          "src/test/rules/de/hypoport/emma/nl/rules/simple-rule-118.drl",
          "src/test/rules/de/hypoport/emma/nl/rules/simple-rule-119.drl",
          "src/test/rules/de/hypoport/emma/nl/rules/simple-rule-120.drl",
          "src/test/rules/de/hypoport/emma/nl/rules/simple-rule-121.drl",
          "src/test/rules/de/hypoport/emma/nl/rules/simple-rule-122.drl"
      }
  )
  public void even_with_25_packages_the_rule_CheckMyDeclaredType_should_still_fire(KnowledgeBase knowledgeBase, String[] packagedRuleFiles) throws Exception {
    runTestForRuleSet(knowledgeBase, packagedRuleFiles, REGULAR_PACKAGE_NAME);
  }

  @Test(dataProviderClass = KnowledgeBaseDataProvider.class, dataProvider = "compileRuleFiles")
  @RuleFiles(
      {
          "src/test/rules/de/hypoport/emma/nl/rules/platform/calculations/declared-type-without-declared-function.drl",
          "src/test/rules/de/hypoport/emma/nl/rules/platform/calculations/check-rule-for-test-using-static-function.drl",
          "src/test/rules/de/hypoport/emma/nl/rules/simple-rule-100.drl",
          "src/test/rules/de/hypoport/emma/nl/rules/simple-rule-101.drl",
          "src/test/rules/de/hypoport/emma/nl/rules/simple-rule-102.drl",
          "src/test/rules/de/hypoport/emma/nl/rules/simple-rule-103.drl",
          "src/test/rules/de/hypoport/emma/nl/rules/simple-rule-104.drl",
          "src/test/rules/de/hypoport/emma/nl/rules/simple-rule-105.drl",
          "src/test/rules/de/hypoport/emma/nl/rules/simple-rule-106.drl",
          "src/test/rules/de/hypoport/emma/nl/rules/simple-rule-107.drl",
          "src/test/rules/de/hypoport/emma/nl/rules/simple-rule-108.drl",
          "src/test/rules/de/hypoport/emma/nl/rules/simple-rule-109.drl",
          "src/test/rules/de/hypoport/emma/nl/rules/simple-rule-110.drl",
          "src/test/rules/de/hypoport/emma/nl/rules/simple-rule-111.drl",
          "src/test/rules/de/hypoport/emma/nl/rules/simple-rule-112.drl",
          "src/test/rules/de/hypoport/emma/nl/rules/simple-rule-113.drl",
          "src/test/rules/de/hypoport/emma/nl/rules/simple-rule-114.drl",
          "src/test/rules/de/hypoport/emma/nl/rules/simple-rule-115.drl",
          "src/test/rules/de/hypoport/emma/nl/rules/simple-rule-116.drl",
          "src/test/rules/de/hypoport/emma/nl/rules/simple-rule-117.drl",
          "src/test/rules/de/hypoport/emma/nl/rules/simple-rule-118.drl",
          "src/test/rules/de/hypoport/emma/nl/rules/simple-rule-119.drl",
          "src/test/rules/de/hypoport/emma/nl/rules/simple-rule-120.drl",
          "src/test/rules/de/hypoport/emma/nl/rules/simple-rule-121.drl",
          "src/test/rules/de/hypoport/emma/nl/rules/simple-rule-122.drl"
      }
  )
  public void even_with_25_packages_the_rule_CheckMyDeclaredType_does_fire_if_using_static_java_function(KnowledgeBase knowledgeBase, String[] packagedRuleFiles) throws Exception {
    runTestForRuleSet(knowledgeBase, packagedRuleFiles, REGULAR_PACKAGE_NAME);
  }


  @Test(dataProviderClass = KnowledgeBaseDataProvider.class, dataProvider = "compileRuleFiles")
  @RuleFiles(
      {
          "src/test/rules/de/hypoport/emma/nl/rules/platform/calculations/declared-type-with-slightly-different-package-name.drl",
          "src/test/rules/de/hypoport/emma/nl/rules/platform/calculations/check-rule-for-test-using-declared-type-from-slightly-different-package.drl",
          "src/test/rules/de/hypoport/emma/nl/rules/simple-rule-100.drl",
          "src/test/rules/de/hypoport/emma/nl/rules/simple-rule-101.drl",
          "src/test/rules/de/hypoport/emma/nl/rules/simple-rule-102.drl",
          "src/test/rules/de/hypoport/emma/nl/rules/simple-rule-103.drl",
          "src/test/rules/de/hypoport/emma/nl/rules/simple-rule-104.drl",
          "src/test/rules/de/hypoport/emma/nl/rules/simple-rule-105.drl",
          "src/test/rules/de/hypoport/emma/nl/rules/simple-rule-106.drl",
          "src/test/rules/de/hypoport/emma/nl/rules/simple-rule-107.drl",
          "src/test/rules/de/hypoport/emma/nl/rules/simple-rule-108.drl",
          "src/test/rules/de/hypoport/emma/nl/rules/simple-rule-109.drl",
          "src/test/rules/de/hypoport/emma/nl/rules/simple-rule-110.drl",
          "src/test/rules/de/hypoport/emma/nl/rules/simple-rule-111.drl",
          "src/test/rules/de/hypoport/emma/nl/rules/simple-rule-112.drl",
          "src/test/rules/de/hypoport/emma/nl/rules/simple-rule-113.drl",
          "src/test/rules/de/hypoport/emma/nl/rules/simple-rule-114.drl",
          "src/test/rules/de/hypoport/emma/nl/rules/simple-rule-115.drl",
          "src/test/rules/de/hypoport/emma/nl/rules/simple-rule-116.drl",
          "src/test/rules/de/hypoport/emma/nl/rules/simple-rule-117.drl",
          "src/test/rules/de/hypoport/emma/nl/rules/simple-rule-118.drl",
          "src/test/rules/de/hypoport/emma/nl/rules/simple-rule-119.drl",
          "src/test/rules/de/hypoport/emma/nl/rules/simple-rule-120.drl",
          "src/test/rules/de/hypoport/emma/nl/rules/simple-rule-121.drl",
          "src/test/rules/de/hypoport/emma/nl/rules/simple-rule-122.drl"
      }
  )
  public void even_with_25_packages_the_rule_CheckMyDeclaredType_does_fire_if_renaming_the_package_slightly(KnowledgeBase knowledgeBase, String[] packagedRuleFiles) throws Exception {
    runTestForRuleSet(knowledgeBase, packagedRuleFiles, SLIGHTLY_DIFFERENT_PACKAGE_NAME);
  }


  private void runTestForRuleSet(KnowledgeBase knowledgeBase, String[] packagedRuleFiles, final String packageName) throws InstantiationException, IllegalAccessException {
    final FactType myDeclaredFactType = knowledgeBase.getFactType(packageName, "MyDeclaredType");
    Object myDeclaredFactInstance = myDeclaredFactType.newInstance();

    StatefulKnowledgeSession knowledgeSession = knowledgeBase.newStatefulKnowledgeSession();
    knowledgeSession.insert(myDeclaredFactInstance);
    knowledgeSession.fireAllRules();

    Collection<Object> matchingFindings = knowledgeSession.getObjects(createMatchingStringFilter("THIS-IS-MY-MARKER-STRING"));

    logComplexityOfKnowledgeBase(knowledgeBase);
    logPackagedRuleFiles(packagedRuleFiles);

    assertThat(matchingFindings).hasSize(1);
    log.info("==== TEST DONE.");
  }

  private ObjectFilter createMatchingStringFilter(final String expectedString) {
    return new ObjectFilter() {
      public boolean accept(Object object) {
        boolean isString = object.getClass().equals(String.class);
        if (!isString) {
          return false;
        }
        String theString = (String) object;
        return expectedString.equals(theString);
      }
    };
  }

  private void logComplexityOfKnowledgeBase(KnowledgeBase knowledgeBase) {
    log.info("KnowledgeBase type: " + knowledgeBase.getClass().getName());
    if (knowledgeBase instanceof org.drools.impl.KnowledgeBaseImpl) {
      KnowledgeBaseImpl knowledgeBaseImpl = (KnowledgeBaseImpl) knowledgeBase;

      ReteooRuleBase ruleBase = (ReteooRuleBase) knowledgeBaseImpl.getRuleBase();
      long numberOfRules = countRules(ruleBase);
      log.info("Node count: " + ruleBase.getNodeCount());
      log.info("Global count: " + ruleBase.getGlobals().size());
      log.info("Package count: " + ruleBase.getPackages().length);
      log.info("Rules count: " + numberOfRules);
      final Rete rete = ruleBase.getRete();
      log.info("Rete complexity: ");
      log.info("   entry point nodes: " + rete.getEntryPointNodes().size());
      log.info("   object type nodes: " + rete.getObjectTypeNodes().size());
    }
  }

  private long countRules(ReteooRuleBase ruleBase) {
    long numberOfRules = 0l;
    for (Package currentPackage : ruleBase.getPackages()) {
      numberOfRules += countRules(currentPackage);
    }
    return numberOfRules;
  }

  private long countRules(Package thePackage) {
    return (thePackage.getRules().length);
  }

  private void logPackagedRuleFiles(String[] packagedRuleFiles) {
    for (String packagedRuleFile : packagedRuleFiles) {
      log.info("File in knowledge base: " + packagedRuleFile);
    }
  }
}
